#!/usr/bin/env python

import argparse, math, sys
from copy import deepcopy

import torch, torchvision

from torch import nn
import torch.nn.functional as F

######################################################################

if torch.cuda.is_available():
    device = torch.device('cuda')
    torch.backends.cudnn.benchmark = True
else:
    device = torch.device('cpu')

######################################################################

parser = argparse.ArgumentParser(
    description = 'An implementation of Mutual Information estimator with a deep model',
    formatter_class = argparse.ArgumentDefaultsHelpFormatter
)

parser.add_argument('--data',
                    type = str, default = 'image_pair',
                    help = 'What data')

parser.add_argument('--seed',
                    type = int, default = 0,
                    help = 'Random seed (default 0, < 0 is no seeding)')

parser.add_argument('--mnist_classes',
                    type = str, default = '0, 1, 3, 5, 6, 7, 8, 9',
                    help = 'What MNIST classes to use')

######################################################################

def entropy(target):
    probas = []
    for k in range(target.max() + 1):
        n = (target == k).sum().item()
        if n > 0: probas.append(n)
    probas = torch.tensor(probas).float()
    probas /= probas.sum()
    return - (probas * probas.log()).sum().item()

######################################################################

args = parser.parse_args()

if args.seed >= 0:
    torch.manual_seed(args.seed)

used_MNIST_classes = torch.tensor(eval('[' + args.mnist_classes + ']'), device = device)

######################################################################

train_set = torchvision.datasets.MNIST('./data/mnist/', train = True, download = True)
train_input  = train_set.train_data.view(-1, 1, 28, 28).to(device).float()
train_target = train_set.train_labels.to(device)

test_set = torchvision.datasets.MNIST('./data/mnist/', train = False, download = True)
test_input = test_set.test_data.view(-1, 1, 28, 28).to(device).float()
test_target = test_set.test_labels.to(device)

mu, std = train_input.mean(), train_input.std()
train_input.sub_(mu).div_(std)
test_input.sub_(mu).div_(std)

######################################################################

# Returns a triplet of tensors (a, b, c), where a and b contain each
# half of the samples, with a[i] and b[i] of same class for any i, and
# c is a 1d long tensor real classes

def create_image_pairs(train = False):
    ua, ub, uc = [], [], []

    if train:
        input, target = train_input, train_target
    else:
        input, target = test_input, test_target

    for i in used_MNIST_classes:
        used_indices = torch.arange(input.size(0), device = target.device)\
                            .masked_select(target == i.item())
        x = input[used_indices]
        x = x[torch.randperm(x.size(0))]
        hs = x.size(0)//2
        ua.append(x.narrow(0, 0, hs))
        ub.append(x.narrow(0, hs, hs))
        uc.append(target[used_indices])

    a = torch.cat(ua, 0)
    b = torch.cat(ub, 0)
    c = torch.cat(uc, 0)
    perm = torch.randperm(a.size(0))
    a = a[perm].contiguous()
    b = b[perm].contiguous()

    return a, b, c

######################################################################

# Returns a triplet a, b, c where a are the standard MNIST images, c
# the classes, and b is a Nx2 tensor, eith for every n:
#
#   b[n, 0] ~ Uniform(0, 10)
#   b[n, 1] ~ b[n, 0] + Uniform(0, 0.5) + c[n]

def create_image_values_pairs(train = False):
    ua, ub = [], []

    if train:
        input, target = train_input, train_target
    else:
        input, target = test_input, test_target

    m = torch.zeros(used_MNIST_classes.max() + 1, dtype = torch.uint8, device = target.device)
    m[used_MNIST_classes] = 1
    m = m[target]
    used_indices = torch.arange(input.size(0), device = target.device).masked_select(m)

    input = input[used_indices].contiguous()
    target = target[used_indices].contiguous()

    a = input
    c = target

    b = a.new(a.size(0), 2)
    b[:, 0].uniform_(10)
    b[:, 1].uniform_(0.5)
    b[:, 1] += b[:, 0] + target.float()

    return a, b, c

######################################################################

def create_sequences_pairs(train = False):
    nb, length = 10000, 1024
    noise_level = 1e-2

    nb_classes = 4
    ha = torch.randint(nb_classes, (nb, ), device = device) + 1
    # hb = torch.randint(nb_classes, (nb, ), device = device)
    hb = ha

    pos = torch.empty(nb, device = device).uniform_(0.0, 0.9)
    a = torch.linspace(0, 1, length, device = device).view(1, -1).expand(nb, -1)
    a = a - pos.view(nb, 1)
    a = (a >= 0).float() * torch.exp(-a * math.log(2) / 0.1)
    a = a * ha.float().view(-1, 1).expand_as(a) / (1 + nb_classes)
    noise = a.new(a.size()).normal_(0, noise_level)
    a = a + noise

    pos = torch.empty(nb, device = device).uniform_(0.5)
    b1 = torch.linspace(0, 1, length, device = device).view(1, -1).expand(nb, -1)
    b1 = b1 - pos.view(nb, 1)
    b1 = (b1 >= 0).float() * torch.exp(-b1 * math.log(2) / 0.1)
    pos = pos + hb.float() / (nb_classes + 1) * 0.5
    b2 = torch.linspace(0, 1, length, device = device).view(1, -1).expand(nb, -1)
    b2 = b2 - pos.view(nb, 1)
    b2 = (b2 >= 0).float() * torch.exp(-b2 * math.log(2) / 0.1)

    b = b1 + b2
    noise = b.new(b.size()).normal_(0, noise_level)
    b = b + noise

    ######################################################################
    # for k in range(10):
        # file = open(f'/tmp/dat{k:02d}', 'w')
        # for i in range(a.size(1)):
            # file.write(f'{a[k, i]:f} {b[k,i]:f}\n')
        # file.close()
    # exit(0)
    ######################################################################

    a = (a - a.mean()) / a.std()
    b = (b - b.mean()) / b.std()

    return a, b, ha

######################################################################

class NetForImagePair(nn.Module):
    def __init__(self):
        super(NetForImagePair, self).__init__()
        self.features_a = nn.Sequential(
            nn.Conv2d(1, 16, kernel_size = 5),
            nn.MaxPool2d(3), nn.ReLU(),
            nn.Conv2d(16, 32, kernel_size = 5),
            nn.MaxPool2d(2), nn.ReLU(),
        )

        self.features_b = nn.Sequential(
            nn.Conv2d(1, 16, kernel_size = 5),
            nn.MaxPool2d(3), nn.ReLU(),
            nn.Conv2d(16, 32, kernel_size = 5),
            nn.MaxPool2d(2), nn.ReLU(),
        )

        self.fully_connected = nn.Sequential(
            nn.Linear(256, 200),
            nn.ReLU(),
            nn.Linear(200, 1)
        )

    def forward(self, a, b):
        a = self.features_a(a).view(a.size(0), -1)
        b = self.features_b(b).view(b.size(0), -1)
        x = torch.cat((a, b), 1)
        return self.fully_connected(x)

######################################################################

class NetForImageValuesPair(nn.Module):
    def __init__(self):
        super(NetForImageValuesPair, self).__init__()
        self.features_a = nn.Sequential(
            nn.Conv2d(1, 16, kernel_size = 5),
            nn.MaxPool2d(3), nn.ReLU(),
            nn.Conv2d(16, 32, kernel_size = 5),
            nn.MaxPool2d(2), nn.ReLU(),
        )

        self.features_b = nn.Sequential(
            nn.Linear(2, 32), nn.ReLU(),
            nn.Linear(32, 32), nn.ReLU(),
            nn.Linear(32, 128), nn.ReLU(),
        )

        self.fully_connected = nn.Sequential(
            nn.Linear(256, 200),
            nn.ReLU(),
            nn.Linear(200, 1)
        )

    def forward(self, a, b):
        a = self.features_a(a).view(a.size(0), -1)
        b = self.features_b(b).view(b.size(0), -1)
        x = torch.cat((a, b), 1)
        return self.fully_connected(x)

######################################################################

class NetForSequencePair(nn.Module):

    def feature_model(self):
        return  nn.Sequential(
            nn.Conv1d(1, self.nc, kernel_size = 5),
            nn.MaxPool1d(2), nn.ReLU(),
            nn.Conv1d(self.nc, self.nc, kernel_size = 5),
            nn.MaxPool1d(2), nn.ReLU(),
            nn.Conv1d(self.nc, self.nc, kernel_size = 5),
            nn.MaxPool1d(2), nn.ReLU(),
            nn.Conv1d(self.nc, self.nc, kernel_size = 5),
            nn.MaxPool1d(2), nn.ReLU(),
            nn.Conv1d(self.nc, self.nc, kernel_size = 5),
            nn.MaxPool1d(2), nn.ReLU(),
        )

    def __init__(self):
        super(NetForSequencePair, self).__init__()

        self.nc = 32
        self.nh = 256

        self.features_a = self.feature_model()
        self.features_b = self.feature_model()

        self.fully_connected = nn.Sequential(
            nn.Linear(2 * self.nc, self.nh),
            nn.ReLU(),
            nn.Linear(self.nh, 1)
        )

    def forward(self, a, b):
        a = a.view(a.size(0), 1, a.size(1))
        a = self.features_a(a)
        a = F.avg_pool1d(a, a.size(2))

        b = b.view(b.size(0), 1, b.size(1))
        b = self.features_b(b)
        b = F.avg_pool1d(b, b.size(2))

        x = torch.cat((a.view(a.size(0), -1), b.view(b.size(0), -1)), 1)
        return self.fully_connected(x)

######################################################################

if args.data == 'image_pair':
    create_pairs = create_image_pairs
    model = NetForImagePair()
elif args.data == 'image_values_pair':
    create_pairs = create_image_values_pairs
    model = NetForImageValuesPair()
elif args.data == 'sequence_pair':
    create_pairs = create_sequences_pairs
    model = NetForSequencePair()
else:
    raise Exception('Unknown data ' + args.data)

######################################################################

nb_epochs, batch_size = 50, 100

print('nb_parameters %d' % sum(x.numel() for x in model.parameters()))

optimizer = torch.optim.Adam(model.parameters(), lr = 1e-3)

model.to(device)

for e in range(nb_epochs):

    input_a, input_b, classes = create_pairs(train = True)

    input_br = input_b[torch.randperm(input_b.size(0))]

    acc_mi = 0.0

    for batch_a, batch_b, batch_br in zip(input_a.split(batch_size),
                                          input_b.split(batch_size),
                                          input_br.split(batch_size)):
        mi = model(batch_a, batch_b).mean() - model(batch_a, batch_br).exp().mean().log()
        loss = - mi
        acc_mi += mi.item()
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

    acc_mi /= (input_a.size(0) // batch_size)

    print('%d %.04f %.04f' % (e + 1, acc_mi / math.log(2), entropy(classes) / math.log(2)))

    sys.stdout.flush()

######################################################################

input_a, input_b, classes = create_pairs(train = False)

for e in range(nb_epochs):
    input_br = input_b[torch.randperm(input_b.size(0))]

    acc_mi = 0.0

    for batch_a, batch_b, batch_br in zip(input_a.split(batch_size),
                                          input_b.split(batch_size),
                                          input_br.split(batch_size)):
        mi = model(batch_a, batch_b).mean() - model(batch_a, batch_br).exp().mean().log()
        acc_mi += mi.item()

    acc_mi /= (input_a.size(0) // batch_size)

print('test %.04f %.04f'%(acc_mi / math.log(2), entropy(classes) / math.log(2)))

######################################################################
